"""Tests CLI utils for generating delivery reports."""

from datetime import datetime

import click.exceptions
import pytest

from cg.cli.generate.delivery_report.utils import (
    get_report_case,
    get_report_api,
    get_report_api_workflow,
    get_report_analysis_started_at,
)
from cg.constants import Workflow
from cg.meta.delivery_report.delivery_report_api import DeliveryReportAPI
from cg.meta.delivery_report.raredisease import RarediseaseDeliveryReportAPI
from cg.store.models import Case


def test_get_report_case(
    raredisease_delivery_report_click_context: click.Context, raredisease_case_id: str
):
    """Tests case object extraction for delivery report generation given a valid case ID."""

    # GIVEN a delivery report click context

    # WHEN resolving a report case ID
    case: Case = get_report_case(
        context=raredisease_delivery_report_click_context, case_id=raredisease_case_id
    )

    # THEN a valid case should be returned
    assert case.internal_id == raredisease_case_id


def test_get_report_case_invalid_case_id(
    raredisease_delivery_report_click_context: click.Context, case_id_does_not_exist: str, caplog
):
    """Tests error raise when an incorrect case ID is provided."""

    # GIVEN a delivery report click context

    # WHEN resolving an invalid report case ID

    # THEN an exception should be raised
    with pytest.raises(click.exceptions.Abort):
        get_report_case(
            context=raredisease_delivery_report_click_context, case_id=case_id_does_not_exist
        )


def test_get_report_api(
    raredisease_delivery_report_click_context: click.Context, raredisease_case_id: str
):
    """Tests delivery report API extraction."""

    # GIVEN a case
    case: Case = get_report_case(
        context=raredisease_delivery_report_click_context, case_id=raredisease_case_id
    )

    # WHEN retrieving a report API
    report_api: DeliveryReportAPI = get_report_api(
        context=raredisease_delivery_report_click_context, case=case
    )

    # THEN the extracted API should be correctly retrieved
    assert report_api
    assert isinstance(report_api, RarediseaseDeliveryReportAPI)


def test_get_report_api_workflow(raredisease_delivery_report_click_context: click.Context):
    """Tests API assignment given a specific workflow."""

    # GIVEN a click context and a specific workflow

    # WHEN validating a report api
    report_api = get_report_api_workflow(
        context=raredisease_delivery_report_click_context, workflow=Workflow.RAREDISEASE
    )

    # THEN the extracted API should match the expected type
    assert report_api
    assert isinstance(report_api, RarediseaseDeliveryReportAPI)


def test_get_report_analysis_started_at(
    raredisease_delivery_report_click_context: click.Context, raredisease_case_id: str
):
    """Tests retrieval of analysis started at field"""

    # GIVEN a case and a report api
    case: Case = get_report_case(
        context=raredisease_delivery_report_click_context, case_id=raredisease_case_id
    )
    report_api: DeliveryReportAPI = get_report_api(
        context=raredisease_delivery_report_click_context, case=case
    )

    # WHEN resolving the analysis started at field
    started_at = get_report_analysis_started_at(
        case=case, report_api=report_api, analysis_started_at=None
    )

    # THEN check if the verified value has been correctly extracted and formatted
    assert started_at
    assert isinstance(started_at, datetime)



from abc import abstractmethod, ABC
from pathlib import Path

from cg.services.deliver_files.file_fetcher.models import SampleFile, CaseFile
from cg.services.deliver_files.file_formatter.destination.models import FormattedFile


class FileFormatter(ABC):

    @abstractmethod
    def format_files(
        self, moved_files: list[CaseFile | SampleFile], delivery_path: Path
    ) -> list[FormattedFile]:
        """Format the files to deliver."""
        pass



"""Tests for the Illumina Sequencing Run filters."""

from sqlalchemy.orm import Query

from cg.constants import SequencingRunDataAvailability
from cg.models.run_devices.illumina_run_directory_data import IlluminaRunDirectoryData
from cg.store.filters.status_illumina_sequencing_run_filters import (
    filter_by_data_availability,
    filter_by_device_internal_id,
    filter_by_entry_id,
)
from cg.store.models import IlluminaSequencingRun
from cg.store.store import Store


def test_filter_by_run_internal_id(
    store_with_illumina_sequencing_data: Store,
    novaseq_x_flow_cell_id: str,
    seven_canonical_flow_cells: list[IlluminaRunDirectoryData],
):
    """Test to filter sequencing runs by run internal id."""
    # GIVEN a store with Illumina Sequencing Runs for canonical flow cells
    sequencing_runs: Query = store_with_illumina_sequencing_data._get_query(
        table=IlluminaSequencingRun
    )
    number_of_flow_cells_in_store: int = len(seven_canonical_flow_cells)
    assert sequencing_runs.count() == number_of_flow_cells_in_store

    # WHEN filtering sequencing runs by run internal id
    filtered_runs: Query = filter_by_device_internal_id(
        runs=sequencing_runs, device_internal_id=novaseq_x_flow_cell_id
    )

    # THEN assert that the filtered query has the correct run
    assert filtered_runs.count() < number_of_flow_cells_in_store
    assert filtered_runs.first().device.internal_id == novaseq_x_flow_cell_id


def test_filter_by_run_entry_id(
    store_with_illumina_sequencing_data: Store,
    seven_canonical_flow_cells: list[IlluminaRunDirectoryData],
):
    """Test to filter sequencing runs by run internal id."""
    # GIVEN a store with Illumina Sequencing Runs for canonical flow cells
    sequencing_runs: Query = store_with_illumina_sequencing_data._get_query(
        table=IlluminaSequencingRun
    )
    number_of_flow_cells_in_store: int = len(seven_canonical_flow_cells)
    assert sequencing_runs.count() == number_of_flow_cells_in_store

    # WHEN filtering sequencing runs by run internal id
    filtered_runs: Query = filter_by_entry_id(runs=sequencing_runs, entry_id=5)

    # THEN assert that the filtered query has the correct run
    assert filtered_runs.count() < number_of_flow_cells_in_store
    assert filtered_runs.first().id == 5


def test_filter_by_data_availability(
    store_with_illumina_sequencing_data: Store,
):
    """Test filter sequencing runs by data availability."""
    # GIVEN a store with Illumina Sequencing Runs with data availability
    sequencing_runs: Query = store_with_illumina_sequencing_data._get_query(
        table=IlluminaSequencingRun
    )
    for sequencing_run in sequencing_runs:
        assert sequencing_run.data_availability

    # WHEN filtering sequencing runs by data availability
    filtered_runs: Query = filter_by_data_availability(
        runs=sequencing_runs, data_availability=[SequencingRunDataAvailability.ON_DISK]
    )

    # THEN the runs with the availability are returned
    for run in filtered_runs:
        assert run.data_availability == SequencingRunDataAvailability.ON_DISK



import pytest

from cg.models.cg_config import CGConfig, PostProcessingServices, RunNamesServices


@pytest.fixture
def pac_bio_context(
    cg_context: CGConfig,
    post_processing_services: PostProcessingServices,
    run_names_services: RunNamesServices,
) -> CGConfig:
    cg_context.post_processing_services_ = post_processing_services
    cg_context.run_names_services_ = run_names_services
    return cg_context



from pathlib import Path
from cg.meta.workflow.mutant.quality_controller.models import (
    MutantPoolSamples,
    MutantQualityResult,
    CaseQualityResult,
    SampleQualityResults,
    SamplesQualityResults,
    ParsedSampleResults,
    SamplePoolAndResults,
)
from cg.meta.workflow.mutant.quality_controller.quality_controller import (
    MutantQualityController,
)
from cg.store.models import Case, Sample


def test_get_mutant_pool_samples(
    mutant_quality_controller: MutantQualityController,
    mutant_case_qc_pass: Case,
    sample_qc_pass: Sample,
    external_negative_control_qc_pass: Sample,
    internal_negative_control_qc_pass: Sample,
):
    # WHEN creating a MutantPoolSamples object
    mutant_pool_samples: MutantPoolSamples = mutant_quality_controller._get_mutant_pool_samples(
        case=mutant_case_qc_pass
    )

    # THEN the pool is created correctly:
    #   - the external negative control is identified and separated from the rest of the samples
    #   - all other samples are present in the list under samples
    #   - the internal negative control corresponding to the case is fetched from lims and added to the pool

    assert mutant_pool_samples.external_negative_control == external_negative_control_qc_pass
    assert mutant_pool_samples.samples == [sample_qc_pass]
    assert mutant_pool_samples.internal_negative_control == internal_negative_control_qc_pass


def test_get_sample_pool_and_results(
    mutant_quality_controller: MutantQualityController,
    mutant_results_file_path_case_qc_pass: Path,
    mutant_case_qc_pass: Case,
    mutant_sample_results_sample_qc_pass: ParsedSampleResults,
    sample_qc_pass: Sample,
):
    # GIVEN a case

    # WHEN generating the quality_metrics
    sample_pool_and_results: SamplePoolAndResults = (
        mutant_quality_controller._get_sample_pool_and_results(
            case_results_file_path=mutant_results_file_path_case_qc_pass,
            case=mutant_case_qc_pass,
        )
    )

    # THEN no errors are raised and the sample_results are created for each sample
    assert (
        sample_pool_and_results.results[sample_qc_pass.internal_id]
        == mutant_sample_results_sample_qc_pass
    )


def test_get_sample_quality_result_for_sample(
    mutant_quality_controller: MutantQualityController,
    sample_qc_pass: Sample,
    mutant_sample_results_sample_qc_pass: ParsedSampleResults,
):
    # GIVEN a sample that passes qc and its corresponding SampleResults

    # WHEN peforming quality control on the sample
    sample_quality_results_sample_qc_pass: SampleQualityResults = (
        mutant_quality_controller._get_sample_quality_result_for_sample(
            sample=sample_qc_pass,
            sample_results=mutant_sample_results_sample_qc_pass,
        )
    )
    # THEN the sample passes qc
    assert sample_quality_results_sample_qc_pass.passes_qc is True


def test_get_sample_quality_result_for_internal_negative_control_sample(
    mutant_quality_controller: MutantQualityController,
    internal_negative_control_qc_pass: Sample,
):
    # GIVEN an internal negative control sample that passes qc and its corresponding SampleResults

    # WHEN peforming quality control on the sample
    sample_quality_results_sample_qc_pass: SampleQualityResults = (
        mutant_quality_controller._get_sample_quality_result_for_internal_negative_control_sample(
            sample=internal_negative_control_qc_pass,
        )
    )
    # THEN the sample passes qc
    assert sample_quality_results_sample_qc_pass.passes_qc is True


def test_get_sample_quality_result_for_external_negative_control_sample(
    mutant_quality_controller: MutantQualityController,
    external_negative_control_qc_pass: Sample,
    mutant_sample_results_external_negative_control_qc_pass: ParsedSampleResults,
):
    # GIVEN an external negative control sample that passes qc and its corresponding SampleResults

    # WHEN peforming quality control on the sample
    sample_quality_results_sample_qc_pass: SampleQualityResults = (
        mutant_quality_controller._get_sample_quality_result_for_external_negative_control_sample(
            sample=external_negative_control_qc_pass,
            sample_results=mutant_sample_results_external_negative_control_qc_pass,
        )
    )
    # THEN the sample passes qc
    assert sample_quality_results_sample_qc_pass.passes_qc is True


def test_get_samples_quality_results(
    mutant_quality_controller: MutantQualityController,
    mutant_sample_pool_and_results_case_qc_pass: SamplePoolAndResults,
):
    # GIVEN a quality metrics objrect from a case where all samples pass QC

    # WHEN performing quality control on all the samples
    samples_quality_results: SamplesQualityResults = (
        mutant_quality_controller._get_samples_quality_results(
            sample_pool_and_results=mutant_sample_pool_and_results_case_qc_pass
        )
    )

    # THEN no error is raised and the correct quality results are generated
    assert samples_quality_results.internal_negative_control.passes_qc is True
    assert samples_quality_results.external_negative_control.passes_qc is True
    assert len(samples_quality_results.samples) == 1
    samples_pass_qc = [
        sample_quality_results.passes_qc
        for sample_quality_results in samples_quality_results.samples
    ]
    assert all(samples_pass_qc) is True


def test_get_case_quality_result(
    mutant_quality_controller: MutantQualityController,
    samples_quality_results_case_qc_pass: SamplesQualityResults,
):
    # GIVEN a samples_quality_results object for a case that passes QC

    # WHEN performing QC on the case
    case_quality_result: CaseQualityResult = mutant_quality_controller._get_case_quality_result(
        samples_quality_results=samples_quality_results_case_qc_pass
    )

    # THEN the correct result is generated
    assert case_quality_result.passes_qc is True
    assert case_quality_result.internal_negative_control_passes_qc is True
    assert case_quality_result.external_negative_control_passes_qc is True


def test_get_quality_control_result_case_qc_pass(
    mutant_quality_controller: MutantQualityController,
    mutant_case_qc_pass: Case,
    mutant_results_file_path_case_qc_pass: Path,
    mutant_qc_report_path_case_qc_pass: Path,
):
    # GIVEN a case that passes QC

    # WHEN performing QC on the case

    case_quality_result: MutantQualityResult = mutant_quality_controller.get_quality_control_result(
        case=mutant_case_qc_pass,
        case_results_file_path=mutant_results_file_path_case_qc_pass,
        case_qc_report_path=mutant_qc_report_path_case_qc_pass,
    )

    # THEN the case passes qc
    assert case_quality_result.passes_qc is True
    assert case_quality_result.case_quality_result.external_negative_control_passes_qc is True
    assert case_quality_result.case_quality_result.internal_negative_control_passes_qc is True


def test_get_quality_control_result_case_qc_fail(
    mutant_quality_controller: MutantQualityController,
    mutant_case_qc_fail: Case,
    mutant_results_file_path_qc_fail: Path,
    mutant_qc_report_path_case_qc_fail: Path,
):
    # GIVEN a case that passes QC

    # WHEN performing QC on the case

    case_quality_result: MutantQualityResult = mutant_quality_controller.get_quality_control_result(
        case=mutant_case_qc_fail,
        case_results_file_path=mutant_results_file_path_qc_fail,
        case_qc_report_path=mutant_qc_report_path_case_qc_fail,
    )

    # THEN the case passes qc
    assert case_quality_result.passes_qc is False
    assert case_quality_result.case_quality_result.external_negative_control_passes_qc is True
    assert case_quality_result.case_quality_result.internal_negative_control_passes_qc is True


def test_get_quality_control_result_case_qc_fail_with_failing_controls(
    mutant_quality_controller: MutantQualityController,
    mutant_case_qc_fail_with_failing_controls: Case,
    mutant_results_file_path_qc_fail_with_failing_controls: Path,
    mutant_qc_report_path_case_qc_fail_with_failing_controls: Path,
):
    # GIVEN a case that does not passe QC due to failing control samples

    # WHEN performing QC on the case

    case_quality_result: MutantQualityResult = mutant_quality_controller.get_quality_control_result(
        case=mutant_case_qc_fail_with_failing_controls,
        case_results_file_path=mutant_results_file_path_qc_fail_with_failing_controls,
        case_qc_report_path=mutant_qc_report_path_case_qc_fail_with_failing_controls,
    )

    # THEN the case does not pass QC and the correct result is retrieved for the control samples
    assert case_quality_result.passes_qc is False
    assert case_quality_result.case_quality_result.external_negative_control_passes_qc is False
    assert case_quality_result.case_quality_result.internal_negative_control_passes_qc is False



"""FOHM report models."""

from pydantic import BaseModel, Field


class FohmComplementaryReport(BaseModel):
    """Model for validating FOHM complementary reports."""

    gisaid_accession: str = Field(str, alias="GISAID_accession")
    internal_id: str | None = Field(None, alias="clinical_genomics_prov_id")
    sample_number: str = Field(str, alias="provnummer")
    selection_criteria: str = Field(str, alias="urvalskriterium")
    region_lab: str | None = Field(None, alias="region_laboratorium")


class FohmPangolinReport(BaseModel):
    """Model for validating FOHM Pangolin reports."""

    ambiguity_score: str
    conflict: str
    constellation_version: str
    internal_id: str | None = None
    is_designated: str
    lineage: str
    note: str
    pangolin_version: str
    qc_notes: str
    region_lab: str | None = None
    qc_status: str
    scorpio_call: str
    scorpio_conflict: str
    scorpio_notes: str
    scorpio_support: str
    scorpio_version: str
    taxon: str
    version: str



"""Module to read or write yaml files"""

from pathlib import Path
from typing import Any

import yaml


def read_yaml(file_path: Path) -> Any:
    """Read content in a yaml file"""
    with open(file_path, "r") as file:
        return yaml.safe_load(file)


def read_yaml_stream(stream: str) -> Any:
    """Read yaml formatted stream"""
    return yaml.safe_load(stream)


def write_yaml(content: Any, file_path: Path) -> None:
    """Write content to a yaml file"""
    with open(file_path, "w") as file:
        yaml.dump(content, file, explicit_start=True)


def write_yaml_stream(content: Any) -> str:
    """Write content to a yaml stream"""
    return yaml.dump(content)


def write_yaml_nextflow_style(content: dict[str, Any], file_path: Path) -> None:
    """Write content to yaml file accepted by Nextflow with non-quoted booleans and quoted strings."""
    with open(file_path, "w") as outfile:
        for key, value in content.items():
            if isinstance(value, Path):
                value: str = value.as_posix()
            quotes = '"' if type(value) is str else ""
            outfile.write(f"{key}: {quotes}{value}{quotes}\n")



import logging
from http import HTTPStatus

from flask import Blueprint, abort, g, jsonify, request

from cg.exc import CaseNotFoundError, CgDataError, OrderMismatchError
from cg.server.dto.delivery_message.delivery_message_request import DeliveryMessageRequest
from cg.server.dto.delivery_message.delivery_message_response import DeliveryMessageResponse
from cg.server.endpoints.utils import before_request
from cg.server.ext import db, delivery_message_service
from cg.store.models import Case, Customer

LOG = logging.getLogger(__name__)
CASES_BLUEPRINT = Blueprint("cases", __name__, url_prefix="/api/v1")
CASES_BLUEPRINT.before_request(before_request)


@CASES_BLUEPRINT.route("/cases")
def get_cases():
    """Return cases with links for a customer from the database."""
    enquiry: str = request.args.get("enquiry")
    action: str = request.args.get("action")

    customers: list[Customer] = _get_current_customers()
    cases: list[Case] = _get_cases(enquiry=enquiry, action=action, customers=customers)

    nr_cases: int = len(cases)
    cases_with_links: list[dict] = [case.to_dict(links=True) for case in cases]
    return jsonify(families=cases_with_links, total=nr_cases)


def _get_current_customers() -> list[Customer] | None:
    """Return customers if the current user is not an admin."""
    return g.current_user.customers if not g.current_user.is_admin else None


def _get_cases(
    enquiry: str | None, action: str | None, customers: list[Customer] | None
) -> list[Case]:
    """Get cases based on the provided filters."""
    return db.get_cases_by_customers_action_and_case_search(
        case_search=enquiry,
        customers=customers,
        action=action,
    )


@CASES_BLUEPRINT.route("/cases/<case_id>")
def parse_case(case_id):
    """Return a case with links."""
    case: Case = db.get_case_by_internal_id(internal_id=case_id)
    if case is None:
        return abort(HTTPStatus.NOT_FOUND)
    if not g.current_user.is_admin and (case.customer not in g.current_user.customers):
        return abort(HTTPStatus.FORBIDDEN)
    return jsonify(**case.to_dict(links=True, analyses=True))


@CASES_BLUEPRINT.route("/cases/delivery_message", methods=["GET"])
def get_cases_delivery_message():
    delivery_message_request = DeliveryMessageRequest.model_validate(request.args)
    try:
        response: DeliveryMessageResponse = delivery_message_service.get_cases_message(
            delivery_message_request
        )
        return jsonify(response.model_dump()), HTTPStatus.OK
    except (CaseNotFoundError, OrderMismatchError, CgDataError) as error:
        return jsonify({"error": str(error)}), HTTPStatus.BAD_REQUEST


@CASES_BLUEPRINT.route("/cases/<case_id>/delivery_message", methods=["GET"])
def get_case_delivery_message(case_id: str):
    delivery_message_request = DeliveryMessageRequest(case_ids=[case_id])
    try:
        response: DeliveryMessageResponse = delivery_message_service.get_cases_message(
            delivery_message_request
        )
        return jsonify(response.model_dump()), HTTPStatus.OK
    except (CaseNotFoundError, CgDataError) as error:
        return jsonify({"error": str(error)}), HTTPStatus.BAD_REQUEST


@CASES_BLUEPRINT.route("/families_in_collaboration")
def get_cases_in_collaboration():
    """Return cases in collaboration."""

    customer_internal_id = request.args.get("customer")
    workflow = request.args.get("data_analysis")
    case_search_pattern = request.args.get("enquiry")

    customer = db.get_customer_by_internal_id(customer_internal_id=customer_internal_id)

    cases = db.get_cases_by_customer_workflow_and_case_search(
        customer=customer, workflow=workflow, case_search=case_search_pattern
    )

    case_dicts = [case.to_dict(links=True) for case in cases]
    return jsonify(families=case_dicts, total=len(cases))


@CASES_BLUEPRINT.route("/families_in_collaboration/<family_id>")
def get_case_in_collaboration(family_id):
    """Return a case with links."""
    case: Case = db.get_case_by_internal_id(internal_id=family_id)
    customer: Customer = db.get_customer_by_internal_id(
        customer_internal_id=request.args.get("customer")
    )
    if case.customer not in customer.collaborators:
        return abort(HTTPStatus.FORBIDDEN)
    return jsonify(**case.to_dict(links=True, analyses=True))



from cg.services.delivery_message.messages.delivery_message import DeliveryMessage
from cg.services.delivery_message.messages.utils import get_caesar_delivery_path
from cg.store.models import Case


class FastqAnalysisMessage(DeliveryMessage):
    def create_message(self, cases: list[Case]) -> str:
        delivery_path: str = get_caesar_delivery_path(cases[0])
        case_names: str = "\n".join([case.name for case in cases])
        case_s_: str = "case" if len(cases) == 1 else "cases"
        return (
            f"Hello,\n\n"
            f"The fastq and analysis files for the following {case_s_} are currently being uploaded to your inbox on Caesar:\n\n"
            f"{case_names}\n\n"
            f"Available under: \n"
            f"{delivery_path}"
        )



"""Schemas for scout serialisation"""

from datetime import datetime

from pydantic import BaseModel, BeforeValidator, ConfigDict, Field
from typing_extensions import Annotated, Literal

from cg.apps.scout.validators import (
    convert_genome_build,
    set_parent_if_missing,
    set_sex_if_other,
)
from cg.constants.gene_panel import GENOME_BUILD_37
from cg.constants.scout import RANK_MODEL_THRESHOLD
from cg.constants.subject import PlinkPhenotypeStatus, PlinkSex, RelationshipStatus, Sex


class Individual(BaseModel):
    model_config = ConfigDict(coerce_numbers_to_str=True)
    bam_file: str | None = None
    individual_id: str
    sex: Annotated[
        Literal[PlinkSex.UNKNOWN, PlinkSex.MALE, PlinkSex.FEMALE, Sex.OTHER],
        BeforeValidator(set_sex_if_other),
    ]
    father: Annotated[str, BeforeValidator(set_parent_if_missing)] = (
        RelationshipStatus.HAS_NO_PARENT
    )
    mother: Annotated[str, BeforeValidator(set_parent_if_missing)] = (
        RelationshipStatus.HAS_NO_PARENT
    )
    phenotype: PlinkPhenotypeStatus
    analysis_type: str = "wgs"


class Panel(BaseModel):
    model_config = ConfigDict(coerce_numbers_to_str=True)
    panel_name: str


class Phenotype(BaseModel):
    model_config = ConfigDict(coerce_numbers_to_str=True)
    phenotype_id: str
    feature: str


class Gene(BaseModel):
    model_config = ConfigDict(coerce_numbers_to_str=True)
    hgnc_id: int
    hgnc_symbol: str | None = None
    region_annotation: str | None = None
    functional_annotation: str | None = None
    sift_prediction: str | None = None
    polyphen_prediction: str | None = None


class DiagnosisPhenotypes(BaseModel):
    model_config = ConfigDict(coerce_numbers_to_str=True)
    disease_nr: int
    disease_id: str
    description: str
    individuals: list[dict[str, str]] | None = None


class ScoutExportCase(BaseModel):
    model_config = ConfigDict(coerce_numbers_to_str=True)
    id: str = Field(str, alias="_id")
    analysis_date: datetime
    owner: str
    causatives: list[str] | None = None
    collaborators: list[str] = []
    individuals: list[Individual]
    genome_build: Annotated[str, BeforeValidator(convert_genome_build)] = GENOME_BUILD_37
    panels: list[Panel] | None = None
    rank_model_version: str | None = None
    sv_rank_model_version: str | None = None
    rank_score_threshold: int = RANK_MODEL_THRESHOLD
    phenotype_terms: list[Phenotype] | None = None
    phenotype_groups: list[Phenotype] | None = None
    diagnosis_phenotypes: list[DiagnosisPhenotypes] | None = None
    diagnosis_genes: list[int] | None = None


class Genotype(BaseModel):
    model_config = ConfigDict(coerce_numbers_to_str=True)
    genotype_call: str
    allele_depths: list[int]
    read_depth: int
    genotype_quality: int
    sample_id: str


class Variant(BaseModel):
    model_config = ConfigDict(coerce_numbers_to_str=True)
    document_id: str
    variant_id: str
    chromosome: str
    position: int
    dbsnp_id: str | None = None
    reference: str
    alternative: str
    quality: float | None = None
    filters: list[str] | None = None
    end: int
    rank_score: int
    category: str
    sub_category: str
    genes: list[Gene] | None = None
    samples: list[Genotype]



from cg.store.models import Case
from cg.store.store import Store


def test_get_uploaded_related_dna_case(
    store_with_rna_and_dna_samples_and_cases: Store,
    rna_case: Case,
    uploaded_related_dna_case: list[Case],
    related_dna_cases: list[Case],
):
    # GIVEN a database with an RNA case and several related DNA cases
    # GIVEN that some of the DNA cases are uploaded and others not
    store: Store = store_with_rna_and_dna_samples_and_cases

    # WHEN getting the related DNA cases that are uploaded
    fetched_uploaded_related_dna_case: list[Case] = store.get_uploaded_related_dna_cases(
        rna_case=rna_case,
    )

    # THEN the correct set of cases is returned
    assert set(fetched_uploaded_related_dna_case) == set(uploaded_related_dna_case)
    assert set(fetched_uploaded_related_dna_case) != set(related_dna_cases)



from pathlib import Path
from cg.meta.workflow.mutant.quality_controller.metrics_parser_utils import (
    _get_sample_name_to_id_mapping,
    _get_samples_results,
    _get_validated_results_list,
    parse_samples_results,
)
from cg.meta.workflow.mutant.quality_controller.models import ParsedSampleResults
from cg.store.models import Case, Sample


def test_get_samples_results(
    mutant_case_qc_pass: Case,
    mutant_results_list_qc_pass: list[ParsedSampleResults],
    sample_qc_pass: Sample,
):
    # GIVEN a case and corresponding results_list

    # WHEN creating a sample_name_to_id_mapping dict
    samples_results: dict[str, ParsedSampleResults] = _get_samples_results(
        case=mutant_case_qc_pass, results_list=mutant_results_list_qc_pass
    )

    # THEN the samples_results object has the correct structure
    assert isinstance(samples_results, dict)
    assert isinstance(samples_results[sample_qc_pass.internal_id], ParsedSampleResults)


def test_get_sample_name_to_id_mapping(mutant_case_qc_pass: Case):
    # GIVEN a case

    # WHEN creating a sample_name_to_id_mapping dict
    sample_name_to_id_mapping: dict[str, str] = _get_sample_name_to_id_mapping(
        case=mutant_case_qc_pass
    )

    # THEN the correct associations are present in the dict
    assert len(sample_name_to_id_mapping) == 2
    assert sample_name_to_id_mapping["sample_qc_pass"] == "sample_qc_pass"
    assert (
        sample_name_to_id_mapping["external_negative_control_qc_pass"]
        == "external_negative_control_qc_pass"
    )


def test_get_validated_results_list(mutant_results_file_path_case_qc_pass: Path):
    # GIVEN a valid raw_results: list[dict[str, Any]] objects

    # WHEN parsing the file
    _get_validated_results_list(results_file_path=mutant_results_file_path_case_qc_pass)

    # THEN no error is thrown


def test_parse_samples_results(
    mutant_case_qc_pass: Case, mutant_results_file_path_case_qc_pass: Path
):
    # GIVEN a case and a valid quality metrics file path

    # WHEN parsing the file
    samples_results: dict[str, ParsedSampleResults] = parse_samples_results(
        case=mutant_case_qc_pass, results_file_path=mutant_results_file_path_case_qc_pass
    )

    # THEN no error is thrown and sample_qc_pass passes QC
    assert samples_results["sample_qc_pass"].passes_qc is True



"""Balsamic QC delivery report API."""

import logging

from cg.meta.delivery_report.balsamic import BalsamicDeliveryReportAPI
from cg.meta.workflow.balsamic_qc import BalsamicQCAnalysisAPI

LOG = logging.getLogger(__name__)


class BalsamicQCDeliveryReportAPI(BalsamicDeliveryReportAPI):
    """API to create Balsamic QC delivery reports."""

    def __init__(self, analysis_api: BalsamicQCAnalysisAPI):
        super().__init__(analysis_api=analysis_api)
        self.analysis_api: BalsamicQCAnalysisAPI = analysis_api



from pathlib import Path

from pydantic import TypeAdapter
from cg.io.csv import read_csv
from typing import Any

from cg.meta.workflow.mutant.quality_controller.models import ParsedSampleResults
from cg.store.models import Case


def parse_samples_results(case: Case, results_file_path: Path) -> dict[str, ParsedSampleResults]:
    """Takes a case object and a results_file_path and resturns dict[str, SampleResults] with sample.internal_id as keys."""

    validated_results_list: list[ParsedSampleResults] = _get_validated_results_list(
        results_file_path=results_file_path
    )

    samples_results: dict[str, ParsedSampleResults] = _get_samples_results(
        case=case, results_list=validated_results_list
    )

    return samples_results


def _get_validated_results_list(results_file_path: Path) -> list[ParsedSampleResults]:
    """Parses the results file and returns a list of validated SampleResults."""
    raw_results: list[dict[Any, Any]] = read_csv(file_path=results_file_path, read_to_dict=True)
    adapter = TypeAdapter(list[ParsedSampleResults])
    return adapter.validate_python(raw_results)


def _get_sample_name_to_id_mapping(case: Case) -> dict[str, str]:
    sample_name_to_id_mapping: dict[str, str] = {}
    for sample in case.samples:
        sample_name_to_id_mapping[sample.name] = sample.internal_id
    return sample_name_to_id_mapping


def _get_samples_results(
    case: Case, results_list: list[ParsedSampleResults]
) -> dict[str, ParsedSampleResults]:
    """Return the mapping of sample internal ids to SampleResults for a case."""

    sample_name_to_id_mapping: dict[str, str] = _get_sample_name_to_id_mapping(case=case)

    samples_results: dict[str, ParsedSampleResults] = {}
    for result in results_list:
        sample_internal_id = sample_name_to_id_mapping[result.sample_name]
        samples_results[sample_internal_id] = result
    return samples_results



"""Constants related to sequencing."""

from enum import StrEnum


class Sequencers(StrEnum):
    """Sequencer instruments."""

    ALL: str = "all"
    HISEQX: str = "hiseqx"
    HISEQGA: str = "hiseqga"
    NOVASEQ: str = "novaseq"
    NOVASEQX: str = "novaseqx"
    OTHER: str = "other"


SEQUENCER_TYPES = {
    "D00134": Sequencers.HISEQGA,
    "D00410": Sequencers.HISEQGA,
    "D00415": Sequencers.HISEQGA,
    "D00450": Sequencers.HISEQGA,
    "D00456": Sequencers.HISEQGA,
    "D00483": Sequencers.HISEQGA,
    "M03284": Sequencers.HISEQGA,
    "SN1025": Sequencers.HISEQGA,
    "SN7001298": Sequencers.HISEQGA,
    "SN7001301": Sequencers.HISEQGA,
    "ST-E00198": Sequencers.HISEQX,
    "ST-E00201": Sequencers.HISEQX,
    "ST-E00214": Sequencers.HISEQX,
    "ST-E00266": Sequencers.HISEQX,
    "ST-E00269": Sequencers.HISEQX,
    "A00187": Sequencers.NOVASEQ,
    "A00621": Sequencers.NOVASEQ,
    "A00689": Sequencers.NOVASEQ,
    "A01901": Sequencers.NOVASEQ,
    "LH00188": Sequencers.NOVASEQX,
    "LH00202": Sequencers.NOVASEQX,
    "LH00217": Sequencers.NOVASEQX,
}

FLOWCELL_Q30_THRESHOLD = {
    Sequencers.HISEQX: 75,
    Sequencers.HISEQGA: 80,
    Sequencers.NOVASEQ: 75,
    Sequencers.NOVASEQX: 75,
}


class Variants(StrEnum):
    """Type of variants."""

    SNV: str = "snv"
    SV: str = "sv"


class RecordType(StrEnum):
    Sample: str = "Sample"
    Pool: str = "Pool"


class SequencingPlatform(StrEnum):
    ILLUMINA: str = "ILLUMINA"
    OXFORD_NANOPORE: str = "OXFORD_NANOPORE"


class SeqLibraryPrepCategory(StrEnum):
    COVID: str = "cov"
    MICROBIAL: str = "mic"
    READY_MADE_LIBRARY: str = "rml"
    TARGETED_GENOME_SEQUENCING: str = "tgs"
    WHOLE_EXOME_SEQUENCING: str = "wes"
    WHOLE_GENOME_SEQUENCING: str = "wgs"
    WHOLE_TRANSCRIPTOME_SEQUENCING: str = "wts"


DNA_PREP_CATEGORIES: list[SeqLibraryPrepCategory] = [
    SeqLibraryPrepCategory.WHOLE_GENOME_SEQUENCING,
    SeqLibraryPrepCategory.TARGETED_GENOME_SEQUENCING,
    SeqLibraryPrepCategory.WHOLE_EXOME_SEQUENCING,
]
